<?php
/**
 * @package   AkeebaLoginGuard
 * @copyright Copyright (c)2016-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

// Prevent direct access
namespace Joomla\Plugin\Loginguard\Yubikey\Extension;

defined('_JEXEC') || die;

use Akeeba\Component\LoginGuard\Administrator\DataShape\CaptiveRenderOptions;
use Akeeba\Component\LoginGuard\Administrator\DataShape\MethodDescriptor;
use Akeeba\Component\LoginGuard\Administrator\DataShape\SetupRenderOptions;
use Akeeba\Component\LoginGuard\Administrator\Event\GetInfo;
use Akeeba\Component\LoginGuard\Administrator\Event\SaveSetup;
use Akeeba\Component\LoginGuard\Administrator\Event\ShowCaptive;
use Akeeba\Component\LoginGuard\Administrator\Event\ShowSetup;
use Akeeba\Component\LoginGuard\Administrator\Event\Validate;
use Akeeba\Component\LoginGuard\Administrator\Helper\Tfa as TfaHelper;
use Akeeba\Component\LoginGuard\Administrator\Table\TfaTable;
use Exception;
use Joomla\CMS\Application\AdministratorApplication;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Application\SiteApplication;
use Joomla\CMS\Http\HttpFactory;
use Joomla\CMS\Input\Input;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\CMS\Uri\Uri;
use Joomla\CMS\User\User;
use Joomla\Event\Event;
use Joomla\Event\SubscriberInterface;
use RuntimeException;

/**
 * Akeeba LoginGuard Plugin for Two Step Verification Method "Yubikey"
 *
 * Use a YubiKey secure hardware token. Supports both the default, centralized key servers and your own custom key
 * server.
 */
class Yubikey extends CMSPlugin implements SubscriberInterface
{
	/**
	 * The application object
	 *
	 * @var    CMSApplication|SiteApplication|AdministratorApplication
	 * @since  6.0.0
	 */
	protected $app;

	/**
	 * Auto-load the plugin's language files
	 *
	 * @var    boolean
	 * @since  6.0.0
	 */
	protected $autoloadLanguage = true;

	/**
	 * The TFA Method name handled by this plugin
	 *
	 * @var    string
	 * @since  1.0.0
	 */
	private $tfaMethodName = 'yubikey';

	/**
	 * Returns an array of events this subscriber will listen to.
	 *
	 * @return  array
	 *
	 * @since   6.0.0
	 */
	public static function getSubscribedEvents(): array
	{
		return [
			'onLoginGuardTfaGetMethod' => 'onLoginGuardTfaGetMethod',
			'onLoginGuardTfaCaptive'   => 'onLoginGuardTfaCaptive',
			'onLoginGuardTfaGetSetup'  => 'onLoginGuardTfaGetSetup',
			'onLoginGuardTfaSaveSetup' => 'onLoginGuardTfaSaveSetup',
			'onLoginGuardTfaValidate'  => 'onLoginGuardTfaValidate',
		];
	}

	/**
	 * Gets the identity of this TFA Method
	 *
	 * @param   GetInfo  $event
	 *
	 * @return  void
	 *
	 * @since   1.0.0
	 */
	public function onLoginGuardTfaGetMethod(GetInfo $event): void
	{
		$event->addResult(
			new MethodDescriptor([
				'name'               => $this->tfaMethodName,
				'display'            => Text::_('PLG_LOGINGUARD_YUBIKEY_LBL_DISPLAYEDAS'),
				'shortinfo'          => Text::_('PLG_LOGINGUARD_YUBIKEY_LBL_SHORTINFO'),
				'image'              => 'media/plg_loginguard_yubikey/images/yubikey.svg',
				'allowMultiple'      => true,
				'help_url'           => $this->params->get('helpurl', 'https://github.com/akeeba/loginguard/wiki/YubiKey'),
				'allowEntryBatching' => 1,
			])
		);
	}

	/**
	 * Returns the information which allows LoginGuard to render the Captive TFA page. This is the page which appears
	 * right after you log in and asks you to validate your login with TFA.
	 *
	 * @param   ShowCaptive  $event
	 *
	 * @return  void
	 * @since   1.0.0
	 */
	public function onLoginGuardTfaCaptive(ShowCaptive $event): void
	{
		/**
		 * @var   TfaTable $record The #__loginguard_tfa record currently selected by the user.
		 */
		$record = $event['record'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->tfaMethodName)
		{
			return;
		}

		$event->addResult(
			new CaptiveRenderOptions([
				// Custom HTML to display above the TFA form
				'pre_message'        => '',
				// How to render the TFA code field. "input" (HTML input element) or "custom" (custom HTML)
				'field_type'         => 'input',
				// The type attribute for the HTML input box. Typically "text" or "password". Use any HTML5 input type.
				'input_type'         => 'text',
				// Placeholder text for the HTML input box. Leave empty if you don't need it.
				'placeholder'        => '',
				// Label to show above the HTML input box. Leave empty if you don't need it.
				'label'              => Text::_('PLG_LOGINGUARD_YUBIKEY_LBL_LABEL'),
				// Custom HTML. Only used when field_type = custom.
				'html'               => '',
				// Custom HTML to display below the TFA form
				'post_message'       => '',
				// URL for help content
				'help_url'           => $this->params->get('helpurl', 'https://github.com/akeeba/loginguard/wiki/YubiKey'),
				// Allow authentication against all entries of this TFA Method. Otherwise authentication takes place against a SPECIFIC entry at a time.
				'allowEntryBatching' => 1,
			])
		);
	}

	/**
	 * Returns the information which allows LoginGuard to render the TFA setup page. This is the page which allows the
	 * user to add or modify a TFA Method for their user account. If the record does not correspond to your plugin
	 * return an empty array.
	 *
	 * @param   ShowSetup  $event
	 *
	 * @return  void
	 * @since   1.0.0
	 */
	public function onLoginGuardTfaGetSetup(ShowSetup $event): void
	{
		/**
		 * @var   TfaTable $record The #__loginguard_tfa record currently selected by the user.
		 */
		$record = $event['record'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->tfaMethodName)
		{
			return;
		}

		// Load the options from the record (if any)
		$options = $this->_decodeRecordOptions($record);
		$keyID   = $options['id'] ?? '';
		$helpURL = $this->params->get('helpurl', 'https://github.com/akeeba/loginguard/wiki/YubiKey');

		if (empty($keyID))
		{
			$event->addResult(
				new SetupRenderOptions([
					'default_title' => Text::_('PLG_LOGINGUARD_YUBIKEY_LBL_DISPLAYEDAS'),
					'pre_message'   => Text::_('PLG_LOGINGUARD_YUBIKEY_LBL_SETUP_INSTRUCTIONS'),
					'field_type'    => 'input',
					'input_type'    => 'text',
					'input_value'   => $keyID,
					'placeholder'   => Text::_('PLG_LOGINGUARD_YUBIKEY_LBL_SETUP_PLACEHOLDER'),
					'label'         => Text::_('PLG_LOGINGUARD_YUBIKEY_LBL_SETUP_LABEL'),
					'help_url'      => $helpURL,
				])
			);
		}
		else
		{
			$event->addResult(
				new SetupRenderOptions([
					'default_title' => Text::_('PLG_LOGINGUARD_YUBIKEY_LBL_DISPLAYEDAS'),
					'pre_message'   => Text::sprintf('PLG_LOGINGUARD_YUBIKEY_LBL_AFTERSETUP_INSTRUCTIONS', $keyID),
					'field_type'    => 'custom',
					'html'          => '',
					'help_url'      => $helpURL,
				])
			);
		}
	}

	/**
	 * Parse the input from the TFA setup page and return the configuration information to be saved to the database. If
	 * the information is invalid throw a RuntimeException to signal the need to display the editor page again. The
	 * message of the exception will be displayed to the user. If the record does not correspond to your plugin return
	 * an empty array.
	 *
	 * @param   SaveSetup  $event
	 *
	 * @return  void
	 *
	 * @throws  Exception
	 * @since   1.0.0
	 */
	public function onLoginGuardTfaSaveSetup(SaveSetup $event): void
	{
		/**
		 * @var   TfaTable $record The #__loginguard_tfa record currently selected by the user.
		 * @var   Input    $input  The user input you are going to take into account.
		 */
		$record = $event['record'];
		$input  = $event['input'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->tfaMethodName)
		{
			return;
		}

		// Load the options from the record (if any)
		$options = $this->_decodeRecordOptions($record);
		$keyID   = $options['id'] ?? '';

		/**
		 * If the submitted code is 12 characters and identical to our existing key there is no change, perform no
		 * further checks.
		 */
		$code = $input->getString('code');

		if ((strlen($code) == 12) && ($code == $keyID))
		{
			$event->addResult($options);

			return;
		}

		// If an empty code or something other than 44 characters was submitted I'm not having any of this!
		if (empty($code) || (strlen($code) != 44))
		{
			throw new RuntimeException(Text::_('PLG_LOGINGUARD_YUBIKEY_ERR_INVALID_CODE'), 500);
		}

		// Validate the code
		$isValid = $this->validateYubikeyOtp($code);

		if (!$isValid)
		{
			throw new RuntimeException(Text::_('PLG_LOGINGUARD_YUBIKEY_ERR_INVALID_CODE'), 500);
		}

		// The code is valid. Keep the Yubikey ID (first twelve characters)
		$keyID = substr($code, 0, 12);

		// Return the configuration to be serialized
		$event->addResult([
			'id' => $keyID,
		]);
	}

	/**
	 * Validates the Two Factor Authentication code submitted by the user in the Captive Two Step Verification page. If
	 * the record does not correspond to your plugin return FALSE.
	 *
	 * @param   Validate  $event
	 *
	 * @return  void
	 * @throws  Exception
	 * @since   1.0.0
	 */
	public function onLoginGuardTfaValidate(Validate $event): void
	{
		/**
		 * @var   TfaTable $record The TFA Method's record you're validatng against
		 * @var   User     $user   The user record
		 * @var   string   $code   The submitted code
		 */
		$record = $event['record'];
		$user   = $event['user'];
		$code   = $event['code'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->tfaMethodName)
		{
			$event->addResult(false);

			return;
		}

		// Double check the TFA Method is for the correct user
		if ($user->id != $record->user_id)
		{
			$event->addResult(false);

			return;
		}

		try
		{
			$records = TfaHelper::getUserTfaRecords($record->user_id);
			$records = array_filter($records, function ($rec) use ($record) {
				return $rec->method === $record->method;
			});
		}
		catch (Exception $e)
		{
			$records = [];
		}

		// Loop all records, stop if at least one matches
		foreach ($records as $aRecord)
		{
			if ($this->validateAgainstRecord($aRecord, $code))
			{
				$event->addResult(true);

				return;
			}
		}

		// None of the records succeeded? Return false.
		$event->addResult(false);
	}

	/**
	 * Validates a Yubikey OTP against the Yubikey servers
	 *
	 * @param   string  $otp  The OTP generated by your Yubikey
	 *
	 * @return  boolean  True if it's a valid OTP
	 * @throws  Exception
	 * @since   1.0.0
	 */
	private function validateYubikeyOtp(string $otp): bool
	{
		// Let the user define a client ID and a secret key in the plugin's configuration
		$clientID     = $this->params->get('client_id', 1);
		$secretKey    = $this->params->get('secret', '');
		$server_queue = trim($this->params->get('servers', ''));

		if (!empty($server_queue))
		{
			$server_queue = explode("\r", $server_queue);
		}

		if (empty($server_queue))
		{
			$server_queue = [
				'https://api.yubico.com/wsapi/2.0/verify',
				'https://api2.yubico.com/wsapi/2.0/verify',
				'https://api3.yubico.com/wsapi/2.0/verify',
				'https://api4.yubico.com/wsapi/2.0/verify',
				'https://api5.yubico.com/wsapi/2.0/verify',
			];
		}

		shuffle($server_queue);

		$gotResponse = false;

		$http     = HttpFactory::getHttp();
		$token    = $this->app->getFormToken();
		$nonce    = md5($token . uniqid(random_int(0, mt_getrandmax())));
		$response = null;

		while (!$gotResponse && !empty($server_queue))
		{
			$server = array_shift($server_queue);
			$uri    = new Uri($server);

			// The client ID for signing the response
			$uri->setVar('id', $clientID);

			// The OTP we read from the user
			$uri->setVar('otp', $otp);

			// This prevents a REPLAYED_OTP status if the token doesn't change after a user submits an invalid OTP
			$uri->setVar('nonce', $nonce);

			// Minimum service level required: 50% (at least 50% of the YubiCloud servers must reply positively for the
			// OTP to validate)
			$uri->setVar('sl', 50);

			// Timeout waiting for YubiCloud servers to reply: 5 seconds.
			$uri->setVar('timeout', 5);

			// Set up the optional HMAC-SHA1 signature for the request.
			$this->signRequest($uri, $secretKey);

			if ($uri->hasVar('h'))
			{
				$uri->setVar('h', urlencode($uri->getVar('h')));
			}

			try
			{
				$response = $http->get($uri->toString(), [], 6);

				if (!empty($response))
				{
					$gotResponse = true;
				}
				else
				{
					continue;
				}
			}
			catch (Exception $exc)
			{
				// No response, continue with the next server
				continue;
			}
		}

		if (empty($response))
		{
			$gotResponse = false;
		}

		// No server replied; we can't validate this OTP
		if (!$gotResponse)
		{
			return false;
		}

		// Parse response
		$lines = explode("\n", $response->body);
		$data  = [];

		foreach ($lines as $line)
		{
			$line  = trim($line);
			$parts = explode('=', $line, 2);

			if (count($parts) < 2)
			{
				continue;
			}

			$data[$parts[0]] = $parts[1];
		}

		// Validate the signature
		$h       = $data['h'] ?? null;
		$fakeUri = Uri::getInstance('http://www.example.com');
		$fakeUri->setQuery($data);
		$this->signRequest($fakeUri, $secretKey);
		$calculatedH = $fakeUri->getVar('h', null);

		if ($calculatedH != $h)
		{
			return false;
		}

		// Validate the response - We need an OK message reply
		if ($data['status'] !== 'OK')
		{
			return false;
		}

		// Validate the response - We need a confidence level over 50%
		if ($data['sl'] < 50)
		{
			return false;
		}

		// Validate the response - The OTP must match
		if ($data['otp'] != $otp)
		{
			return false;
		}

		// Validate the response - The token must match
		if ($data['nonce'] != $nonce)
		{
			return false;
		}

		return true;
	}

	/**
	 * Sign the request to YubiCloud.
	 *
	 * @param   Uri     $uri     The request URI to sign
	 * @param   string  $secret  The secret key to sign with
	 *
	 * @return  void
	 * @since   1.0.0
	 *
	 * @see     https://developers.yubico.com/yubikey-val/Validation_Protocol_V2.0.html
	 */
	private function signRequest(Uri $uri, string $secret): void
	{
		// Make sure we have an encoding secret
		$secret = trim($secret);

		if (empty($secret))
		{
			return;
		}

		// I will need base64 encoding and decoding
		if (!function_exists('base64_encode') || !function_exists('base64_decode'))
		{
			return;
		}

		// I need HMAC-SHA-1 support. Therefore I check for HMAC and SHA1 support in the PHP 'hash' extension.
		if (!function_exists('hash_hmac') || !function_exists('hash_algos'))
		{
			return;
		}

		$algos = hash_algos();

		if (!in_array('sha1', $algos))
		{
			return;
		}

		// Get the parameters
		/** @var   array $vars I have to explicitly state the type because the Joomla docblock is wrong :( */
		$vars = $uri->getQuery(true);

		// 'h' is the hash and it doesn't participate in the calculation of itself.
		if (isset($vars['h']))
		{
			unset($vars['h']);
		}

		// Alphabetically sort the set of key/value pairs by key order.
		ksort($vars);

		/**
		 * Construct a single line with each ordered key/value pair concatenated using &, and each key and value
		 * concatenated with =. Do not add any line breaks. Do not add whitespace.
		 *
		 * Now, if you thought I can't really write PHP code, a.k.a. why not use http_build_query, read on.
		 *
		 * The way YubiKey expects the query to be built is UTTERLY WRONG. They are doing string concatenation, not
		 * URL query building! Therefore you cannot use http_build_query(). Instead, you need to use dumb string
		 * concatenation. I kid you not. If you want to laugh (or cry) read their Auth_Yubico class. It's 1998 all over
		 * again.
		 */
		$stringToSign = '';

		foreach ($vars as $k => $v)
		{
			$stringToSign .= '&' . $k . '=' . $v;
		}

		$stringToSign = ltrim($stringToSign, '&');

		/**
		 * Apply the HMAC-SHA-1 algorithm on the line as an octet string using the API key as key (remember to
		 * base64decode the API key obtained from Yubico).
		 */
		$decodedKey = base64_decode($secret);
		$hash       = hash_hmac('sha1', $stringToSign, $decodedKey, true);

		/**
		 * Base 64 encode the resulting value according to RFC 4648, for example, t2ZMtKeValdA+H0jVpj3LIichn4=
		 */
		$h = base64_encode($hash);

		/**
		 * Append the value under key h to the message.
		 */
		$uri->setVar('h', $h);
	}

	/**
	 * Decodes the options from a #__loginguard_tfa record into an options object.
	 *
	 * @param   TfaTable  $record
	 *
	 * @return  array
	 * @since   1.0.0
	 */
	private function _decodeRecordOptions(TfaTable $record): array
	{
		$options = [
			'id' => '',
		];

		if (!empty($record->options))
		{
			$recordOptions = $record->options;

			$options = array_merge($options, $recordOptions);
		}

		return $options;
	}

	/**
	 * @param   TfaTable  $record
	 * @param   string    $code
	 *
	 * @return  bool
	 * @throws  Exception
	 * @since   1.0.0
	 */
	private function validateAgainstRecord(TfaTable $record, string $code): bool
	{
		// Load the options from the record (if any)
		$options = $this->_decodeRecordOptions($record);
		$keyID   = $options['id'] ?? '';

		// If there is no key in the options throw an error
		if (empty($keyID))
		{
			return false;
		}

		// If the submitted code is empty throw an error
		if (empty($code))
		{
			return false;
		}

		// If the submitted code length is wrong throw an error
		if (strlen($code) != 44)
		{
			return false;
		}

		// If the submitted code's key ID does not match the stored throw an error
		if (substr($code, 0, 12) != $keyID)
		{
			return false;
		}

		// Check the OTP code for validity
		return $this->validateYubikeyOtp($code);
	}
}
